@@ -16,17 +16,24 @@ MANIFEST.in to be included nevertheless.
 from __future__ import print_function
 
 import argparse
+import codecs
 import fnmatch
 import locale
 import os
+import posixpath
 import re
 import shutil
+import stat
 import subprocess
 import sys
 import tarfile
 import tempfile
+import unicodedata
 import zipfile
+from distutils.filelist import glob_to_re
 from contextlib import contextmanager, closing
+from distutils.text_file import TextFile
+from xml.etree import cElementTree as ET
 
 try:
     import ConfigParser
@@ -35,7 +42,7 @@ except ImportError:
     import configparser as ConfigParser
 
 
-__version__ = '0.22'
+__version__ = '0.36'
 __author__ = 'Marius Gedminas <marius@gedmin.as>'
 __licence__ = 'MIT'
 __url__ = 'https://github.com/mgedmin/check-manifest'
@@ -126,7 +133,7 @@ class CommandFailed(Failure):
                                command, status, output))
 
 
-def run(command, encoding=None):
+def run(command, encoding=None, decode=True, cwd=None):
     """Run a command [cmd, arg1, arg2, ...].
 
     Returns the output (stdout + stderr).
@@ -136,11 +143,15 @@ def run(command, encoding=None):
     if not encoding:
         encoding = locale.getpreferredencoding()
     try:
-        pipe = subprocess.Popen(command, stdout=subprocess.PIPE,
-                                stderr=subprocess.STDOUT)
+        with open(os.devnull, 'rb') as devnull:
+            pipe = subprocess.Popen(command, stdin=devnull,
+                                    stdout=subprocess.PIPE,
+                                    stderr=subprocess.STDOUT, cwd=cwd)
     except OSError as e:
         raise Failure("could not run %s: %s" % (command, e))
-    output = pipe.communicate()[0].decode(encoding)
+    output = pipe.communicate()[0]
+    if decode:
+        output = output.decode(encoding)
     status = pipe.wait()
     if status != 0:
         raise CommandFailed(command, status, output)
@@ -171,7 +182,38 @@ def mkdtemp(hint=''):
     try:
         yield dirname
     finally:
-        shutil.rmtree(dirname)
+        rmtree(dirname)
+
+
+def chmod_plus(path, add_bits=stat.S_IWUSR):
+    """Change a file's mode by adding a few bits.
+
+    Like chmod +<bits> <path> in a Unix shell.
+    """
+    try:
+        os.chmod(path, stat.S_IMODE(os.stat(path).st_mode) | add_bits)
+    except OSError:  # pragma: nocover
+        pass  # well, we tried
+
+
+def rmtree(path):
+    """A version of rmtree that can deal with read-only files and directories.
+
+    Needed because the stock shutil.rmtree() fails with an access error
+    when there are read-only files in the directory on Windows, or when the
+    directory itself is read-only on Unix.
+    """
+    def onerror(func, path, exc_info):
+        # Did you know what on Python 3.3 on Windows os.remove() and
+        # os.unlink() are distinct functions?
+        if func is os.remove or func is os.unlink or func is os.rmdir:
+            if sys.platform != 'win32':
+                chmod_plus(os.path.dirname(path), stat.S_IWUSR | stat.S_IXUSR)
+            chmod_plus(path)
+            func(path)
+        else:
+            raise
+    shutil.rmtree(path, onerror=onerror)
 
 
 def copy_files(filelist, destdir):
@@ -259,6 +301,17 @@ def strip_toplevel_name(filelist):
     return [name[len(prefix):] for name in names]
 
 
+def add_prefix_to_each(prefix, filelist):
+    """Add a prefix to each name in a file list.
+
+        >>> filelist = add_prefix_to_each('foo/bar', ['a', 'b', 'c/d'])
+        >>> [f.replace(os.path.sep, '/') for f in filelist]
+        ['foo/bar/a', 'foo/bar/b', 'foo/bar/c/d']
+
+    """
+    return [posixpath.join(prefix, name) for name in filelist]
+
+
 class VCS(object):
 
     @classmethod
@@ -269,14 +322,42 @@ class VCS(object):
 class Git(VCS):
     metadata_name = '.git'
 
-    @staticmethod
-    def get_versioned_files():
+    # Git for Windows uses UTF-8 instead of the locale encoding.
+    # Git on POSIX systems uses the locale encoding.
+    _encoding = 'UTF-8' if sys.platform == 'win32' else None
+
+    @classmethod
+    def detect(cls, location):
+        # .git can be a file for submodules
+        return os.path.exists(os.path.join(location, cls.metadata_name))
+
+    @classmethod
+    def get_versioned_files(cls):
         """List all files versioned by git in the current directory."""
-        # Git for Windows uses UTF-8 instead of the locale encoding.
-        # Regular Git on sane POSIX systems uses the locale encoding
-        encoding = 'UTF-8' if sys.platform == 'win32' else None
-        output = run(['git', 'ls-files', '-z'], encoding=encoding)
-        return add_directories(output.split('\0')[:-1])
+        files = cls._git_ls_files()
+        submodules = cls._list_submodules()
+        for subdir in submodules:
+            subdir = os.path.relpath(subdir).replace(os.path.sep, '/')
+            files += add_prefix_to_each(subdir, cls._git_ls_files(subdir))
+        return add_directories(files)
+
+    @classmethod
+    def _git_ls_files(cls, cwd=None):
+        output = run(['git', 'ls-files', '-z'], encoding=cls._encoding, cwd=cwd)
+        return output.split('\0')[:-1]
+
+    @classmethod
+    def _list_submodules(cls):
+        # This is incredibly expensive on my Jenkins instance (50 seconds for
+        # each invocation, even when there are no submodules whatsoever).
+        # Curiously, I cannot reproduce that in Appveyor, or even on the same
+        # Jenkins machine but when I run the tests manually.  Still, 2-hour
+        # Jenkins runs are bad, so let's avoid running 'git submodule' when
+        # there's no .gitmodules file.
+        if not os.path.exists('.gitmodules'):
+            return []
+        return run(['git', 'submodule', '--quiet', 'foreach', '--recursive',
+                    'printf "%s/%s\\n" $toplevel $path'], encoding=cls._encoding).splitlines()
 
 
 class Mercurial(VCS):
@@ -285,29 +366,98 @@ class Mercurial(VCS):
     @staticmethod
     def get_versioned_files():
         """List all files under Mercurial control in the current directory."""
-        output = run(['hg', 'status', '-ncam', '.'])
+        output = run(['hg', 'status', '-ncamd', '.'])
         return add_directories(output.splitlines())
 
 
 class Bazaar(VCS):
     metadata_name = '.bzr'
 
-    @staticmethod
-    def get_versioned_files():
+    @classmethod
+    def _get_terminal_encoding(self):
+        # Python 3.6 lets us name the OEM codepage directly, which is lucky
+        # because it also breaks our old method of OEM codepage detection
+        # (PEP-528 changed sys.stdout.encoding to UTF-8).
+        try:
+            codecs.lookup('oem')
+        except LookupError:
+            pass
+        else:  # pragma: nocover
+            return 'oem'
+        # Based on bzrlib.osutils.get_terminal_encoding()
+        encoding = getattr(sys.stdout, 'encoding', None)
+        if not encoding:
+            encoding = getattr(sys.stdin, 'encoding', None)
+        if encoding == 'cp0':  # "no codepage"
+            encoding = None
+        # NB: bzrlib falls back on bzrlib.osutils.get_user_encoding(),
+        # which is like locale.getpreferredencoding() on steroids, and
+        # also includes a fallback from 'ascii' to 'utf-8' when
+        # sys.platform is 'darwin'.  This is probably something we might
+        # want to do in run(), but I'll wait for somebody to complain
+        # first, since I don't have a Mac OS X machine and cannot test.
+        return encoding
+
+    @classmethod
+    def get_versioned_files(cls):
         """List all files versioned in Bazaar in the current directory."""
-        output = run(['bzr', 'ls', '-VR'])
+        encoding = cls._get_terminal_encoding()
+        output = run(['bzr', 'ls', '-VR'], encoding=encoding)
         return output.splitlines()
 
 
 class Subversion(VCS):
     metadata_name = '.svn'
 
-    @staticmethod
-    def get_versioned_files():
+    @classmethod
+    def get_versioned_files(cls):
         """List all files under SVN control in the current directory."""
-        output = run(['svn', 'st', '-vq'])
-        return sorted(line[12:].split(None, 3)[-1]
-                      for line in output.splitlines()[1:])
+        output = run(['svn', 'st', '-vq', '--xml'], decode=False)
+        tree = ET.XML(output)
+        return sorted(entry.get('path') for entry in tree.findall('.//entry')
+                      if cls.is_interesting(entry))
+
+    @staticmethod
+    def is_interesting(entry):
+        """Is this entry interesting?
+
+        ``entry`` is an XML node representing one entry of the svn status
+        XML output.  It looks like this::
+
+            <entry path="unchanged.txt">
+              <wc-status item="normal" revision="1" props="none">
+                <commit revision="1">
+                  <author>mg</author>
+                  <date>2015-02-06T07:52:38.163516Z</date>
+                </commit>
+              </wc-status>
+            </entry>
+
+            <entry path="added-but-not-committed.txt">
+              <wc-status item="added" revision="-1" props="none"></wc-status>
+            </entry>
+
+            <entry path="ext">
+              <wc-status item="external" props="none"></wc-status>
+            </entry>
+
+            <entry path="unknown.txt">
+              <wc-status props="none" item="unversioned"></wc-status>
+            </entry>
+
+        """
+        if entry.get('path') == '.':
+            return False
+        status = entry.find('wc-status')
+        if status is None:
+            warning('svn status --xml parse error: <entry path="%s"> without'
+                    ' <wc-status>' % entry.get('path'))
+            return False
+        # For SVN externals we get two entries: one mentioning the
+        # existence of the external, and one about the status of the external.
+        if status.get('item') in ('unversioned', 'external'):
+            return False
+        return True
 
 
 def detect_vcs():
@@ -331,11 +481,25 @@ def get_vcs_files():
 
 
 def normalize_names(names):
+    """Normalize file names."""
+    return list(map(normalize_name, names))
+
+
+def normalize_name(name):
     """Some VCS print directory names with trailing slashes.  Strip them.
 
     Easiest is to normalize the path.
+
+    And encodings may trip us up too, especially when comparing lists
+    of files.  Plus maybe lowercase versus uppercase.
     """
-    return [os.path.normpath(name) for name in names]
+    name = os.path.normpath(name)
+    name = unicodify(name)
+    if sys.platform == 'darwin':
+        # Mac OSX may have problems comparing non-ascii filenames, so
+        # we convert them.
+        name = unicodedata.normalize('NFC', name)
+    return name
 
 
 def add_directories(names):
@@ -363,8 +527,10 @@ IGNORE = [
     '*.egg-info/*', # always generated
     'setup.cfg',    # always generated, sometimes also kept in source control
     # it's not a problem if the sdist is lacking these files:
-    '.hgtags', '.hgignore', '.gitignore', '.bzrignore',
+    '.hgtags', '.hgsigs', '.hgignore', '.gitignore', '.bzrignore',
+    '.gitattributes',
     '.travis.yml',
+    'Jenkinsfile',
     # it's convenient to ship compiled .mo files in sdists, but they shouldn't
     # be checked in
     '*.mo',
@@ -389,6 +555,8 @@ WARN_ABOUT_FILES_IN_VCS = [
     '.#*',
 ]
 
+IGNORE_BAD_IDEAS = []
+
 _sep = r'\\' if os.path.sep == '\\' else os.path.sep
 
 SUGGESTIONS = [(re.compile(pattern.replace('/', _sep)), suggestion) for pattern, suggestion in [
@@ -412,21 +580,29 @@ SUGGESTIONS = [(re.compile(pattern.replace('/', _sep)), suggestion) for pattern,
                                     r'recursive-include \1 *.\2'),
 ]]
 
+CFG_SECTION_CHECK_MANIFEST = 'check-manifest'
+CFG_IGNORE_DEFAULT_RULES = (CFG_SECTION_CHECK_MANIFEST, 'ignore-default-rules')
+CFG_IGNORE = (CFG_SECTION_CHECK_MANIFEST, 'ignore')
+CFG_IGNORE_BAD_IDEAS = (CFG_SECTION_CHECK_MANIFEST, 'ignore-bad-ideas')
+
 
 def read_config():
     """Read configuration from setup.cfg."""
     # XXX modifies global state, which is kind of evil
     config = ConfigParser.ConfigParser()
     config.read(['setup.cfg'])
-    if not config.has_section('check-manifest'):
+    if not config.has_section(CFG_SECTION_CHECK_MANIFEST):
         return
-    if (config.has_option('check-manifest', 'ignore-default-rules')
-            and config.getboolean('check-manifest', 'ignore-default-rules')):
+    if (config.has_option(*CFG_IGNORE_DEFAULT_RULES)
+            and config.getboolean(*CFG_IGNORE_DEFAULT_RULES)):
         del IGNORE[:]
-    if config.has_option('check-manifest', 'ignore'):
-        patterns = [p.strip() for p in config.get('check-manifest',
-                                                  'ignore').splitlines()]
+    if config.has_option(*CFG_IGNORE):
+        patterns = [p.strip() for p in config.get(*CFG_IGNORE).splitlines()]
         IGNORE.extend(p for p in patterns if p)
+    if config.has_option(*CFG_IGNORE_BAD_IDEAS):
+        lines = config.get(*CFG_IGNORE_BAD_IDEAS).splitlines()
+        patterns = [p.strip() for p in lines]
+        IGNORE_BAD_IDEAS.extend(p for p in patterns if p)
 
 
 def read_manifest():
@@ -437,9 +613,7 @@ def read_manifest():
     # XXX modifies global state, which is kind of evil
     if not os.path.isfile('MANIFEST.in'):
         return
-    with open('MANIFEST.in') as manifest:
-        contents = manifest.read()
-    ignore, ignore_regexps = _get_ignore_from_manifest(contents)
+    ignore, ignore_regexps = _get_ignore_from_manifest('MANIFEST.in')
     IGNORE.extend(ignore)
     IGNORE_REGEXPS.extend(ignore_regexps)
 
@@ -448,28 +622,53 @@ def _glob_to_regexp(pat):
     """Compile a glob pattern into a regexp.
 
     We need to do this because fnmatch allows * to match /, which we
-    don't want.  E.g. an MANIFEST.in exclude of 'dirname/*css' should
+    don't want.  E.g. a MANIFEST.in exclude of 'dirname/*css' should
     match 'dirname/foo.css' but not 'dirname/subdir/bar.css'.
     """
-    pat = fnmatch.translate(pat)
-    # Note that distutils in Python 2.6 has a buggy glob_to_re in
-    # distutils.filelist -- it converts '*.cfg' to '[^/]*cfg' instead
-    # of '[^\\]*cfg' on Windows.
-    sep = r'\\\\' if os.path.sep == '\\' else os.path.sep
-    return re.sub(r'((?<!\\)(\\\\)*)\.', r'\1[^%s]' % sep, pat)
+    return glob_to_re(pat)
 
 
-def _get_ignore_from_manifest(contents):
-    """Gather the various ignore patterns from MANIFEST.in.
+def _get_ignore_from_manifest(filename):
+    """Gather the various ignore patterns from a MANIFEST.in.
 
-    'contents' should be a string, which may contain newlines.
+    Returns a list of standard ignore patterns and a list of regular
+    expressions to ignore.
+    """
+
+    class MyTextFile(TextFile):
+        def error(self, msg, line=None):  # pragma: nocover
+            # (this is never called by TextFile in current versions of CPython)
+            raise Failure(self.gen_error(msg, line))
+
+        def warn(self, msg, line=None):
+            warning(self.gen_error(msg, line))
+
+    template = MyTextFile(filename,
+                          strip_comments=True,
+                          skip_blanks=True,
+                          join_lines=True,
+                          lstrip_ws=True,
+                          rstrip_ws=True,
+                          collapse_join=True)
+    try:
+        lines = template.readlines()
+    finally:
+        template.close()
+    return _get_ignore_from_manifest_lines(lines)
+
+
+def _get_ignore_from_manifest_lines(lines):
+    """Gather the various ignore patterns from a MANIFEST.in.
+
+    'lines' should be a list of strings with comments removed
+    and continuation lines joined.
 
     Returns a list of standard ignore patterns and a list of regular
     expressions to ignore.
     """
     ignore = []
     ignore_regexps = []
-    for line in contents.splitlines():
+    for line in lines:
         try:
             cmd, rest = line.split(None, 1)
         except ValueError:
@@ -533,7 +732,9 @@ def _get_ignore_from_manifest(contents):
 
 def file_matches(filename, patterns):
     """Does this filename match any of the patterns?"""
-    return any(fnmatch.fnmatch(filename, pat) for pat in patterns)
+    return any(fnmatch.fnmatch(filename, pat) or
+               fnmatch.fnmatch(os.path.basename(filename), pat)
+               for pat in patterns)
 
 
 def file_matches_regexps(filename, patterns):
@@ -594,7 +795,8 @@ def check_manifest(source_tree='.', create=False, update=False,
     Returns True if the manifest is fine.
     """
     all_ok = True
-    python = os.path.abspath(python)  # in case it was relative
+    if os.path.sep in python:
+        python = os.path.abspath(python)
     with cd(source_tree):
         if not is_package():
             raise Failure('This is not a Python project (no setup.py).')
@@ -604,6 +806,8 @@ def check_manifest(source_tree='.', create=False, update=False,
         all_source_files = sorted(get_vcs_files())
         source_files = strip_sdist_extras(all_source_files)
         info_continue(": %d files and directories" % len(source_files))
+        if not all_source_files:
+            raise Failure('There are no files added to version control!')
         info_begin("building an sdist")
         with mkdtemp('-sdist') as tempdir:
             run([python, 'setup.py', 'sdist', '-d', tempdir])
@@ -627,6 +831,12 @@ def check_manifest(source_tree='.', create=False, update=False,
                 # gets confused about their new manifest rules being
                 # ignored.
                 copy_files(['MANIFEST.in'], tempsourcedir)
+            if 'setup.py' not in source_files:
+                # See https://github.com/mgedmin/check-manifest/issues/46
+                # if do this, we will emit a warning about setup.py not
+                # being in source control, if we don't do this, the user
+                # gets a scary error
+                copy_files(['setup.py'], tempsourcedir)
             info_begin("building a clean sdist")
             with cd(tempsourcedir):
                 with mkdtemp('-sdist') as tempdir:
@@ -672,13 +882,15 @@ def check_manifest(source_tree='.', create=False, update=False,
                      " matching any of the files, sorry!")
             all_ok = False
         bad_ideas = find_bad_ideas(all_source_files)
-        if bad_ideas:
+        filtered_bad_ideas = [bad_idea for bad_idea in bad_ideas
+                              if not file_matches(bad_idea, IGNORE_BAD_IDEAS)]
+        if filtered_bad_ideas:
             warning("you have %s in source control!\nthat's a bad idea:"
                     " auto-generated files should not be versioned"
-                    % bad_ideas[0])
-            if len(bad_ideas) > 1:
+                    % filtered_bad_ideas[0])
+            if len(filtered_bad_ideas) > 1:
                 warning("this also applies to the following:\n%s"
-                        % format_list(bad_ideas[1:]))
+                        % format_list(filtered_bad_ideas[1:]))
             all_ok = False
     return all_ok
 
@@ -706,11 +918,17 @@ def main():
     parser.add_argument('--ignore', metavar='patterns', default=None,
                         help='ignore files/directories matching these'
                              ' comma-separated patterns')
+    parser.add_argument('--ignore-bad-ideas', metavar='patterns',
+                        default=[], help='ignore bad idea files/directories '
+                        'matching these comma-separated patterns')
     args = parser.parse_args()
 
     if args.ignore:
         IGNORE.extend(args.ignore.split(','))
 
+    if args.ignore_bad_ideas:
+        IGNORE_BAD_IDEAS.extend(args.ignore_bad_ideas.split(','))
+
     if args.verbose:
         global VERBOSE
         VERBOSE = True
@@ -732,7 +950,7 @@ def zest_releaser_check(data):
     """Check the completeness of MANIFEST.in before the release.
 
     This is an entry point for zest.releaser.  See the documentation at
-    http://zestreleaser.readthedocs.org/en/latest/entrypoints.html
+    https://zestreleaser.readthedocs.io/en/latest/entrypoints.html
     """
     from zest.releaser.utils import ask
     source_tree = data['workingdir']
@@ -745,7 +963,7 @@ def zest_releaser_check(data):
         return
     try:
         if not check_manifest(source_tree):
-            if not ask("MANIFEST.in is not in order. "
+            if not ask("MANIFEST.in has problems. "
                        " Do you want to continue despite that?", default=False):
                 sys.exit(1)
     except Failure as e:
