@@ -1,20 +1,17 @@
+import codecs
 import doctest
 import locale
 import os
-import shutil
 import subprocess
 import sys
 import tarfile
 import tempfile
 import textwrap
+import unittest
 import zipfile
 from contextlib import closing
 from io import BytesIO
-
-try:
-    import unittest2 as unittest    # Python 2.6
-except ImportError:
-    import unittest
+from xml.etree import cElementTree as ET
 
 try:
     from cStringIO import StringIO  # Python 2.x
@@ -23,22 +20,19 @@ except ImportError:
 
 import mock
 
+from check_manifest import rmtree
 
-def rmtree(path):
-    """A version of rmtree that can remove read-only files on Windows.
 
-    Needed because the stock shutil.rmtree() fails with an access error
-    due to the files in .git/objects being read-only.
-    """
-    def onerror(func, path, exc_info):
-        if func is os.remove or func is os.unlink:
-            # Did you know what on Python 3.3 on Windows os.remove() and
-            # os.unlink() are distinct functions?
-            os.chmod(path, 0o644)
-            func(path)
-        else:
-            raise
-    shutil.rmtree(path, onerror=onerror)
+CAN_SKIP_TESTS = os.getenv('SKIP_NO_TESTS', '') == ''
+
+
+try:
+    codecs.lookup('oem')
+except LookupError:
+    HAS_OEM_CODEC = False
+else:
+    # Python >= 3.6 on Windows
+    HAS_OEM_CODEC = True
 
 
 class Tests(unittest.TestCase):
@@ -58,6 +52,10 @@ class Tests(unittest.TestCase):
         self.addCleanup(rmtree, tmpdir)
         return tmpdir
 
+    def create_file(self, filename, contents):
+        with open(filename, 'w') as f:
+            f.write(contents)
+
     def create_zip_file(self, filename, filenames):
         with closing(zipfile.ZipFile(filename, 'w')) as zf:
             for fn in filenames:
@@ -92,6 +90,49 @@ class Tests(unittest.TestCase):
             '\n%r does not start with\n%r' % (str(cm.exception),
                                               should_start_with))
 
+    def test_mkdtemp_readonly_files(self):
+        from check_manifest import mkdtemp
+        with mkdtemp(hint='-test-readonly') as d:
+            fn = os.path.join(d, 'file.txt')
+            with open(fn, 'w'):
+                pass
+            os.chmod(fn, 0o444)  # readonly
+        assert not os.path.exists(d)
+
+    @unittest.skipIf(sys.platform == 'win32',
+                     "No POSIX-like unreadable directories on Windows")
+    def test_rmtree_unreadable_directories(self):
+        d = self.make_temp_dir()
+        sd = os.path.join(d, 'subdir')
+        os.mkdir(sd)
+        os.chmod(sd, 0)  # a bad mode for a directory, oops
+        # The onerror API of shutil.rmtree doesn't let us recover from
+        # os.listdir() failures.
+        with self.assertRaises(OSError):
+            rmtree(sd)
+        os.chmod(sd, 0o755)  # so we can clean up
+
+    def test_rmtree_readonly_directories(self):
+        d = self.make_temp_dir()
+        sd = os.path.join(d, 'subdir')
+        fn = os.path.join(sd, 'file.txt')
+        os.mkdir(sd)
+        open(fn, 'w').close()
+        os.chmod(sd, 0o444)  # a bad mode for a directory, oops
+        rmtree(sd)
+        assert not os.path.exists(sd)
+
+    def test_rmtree_readonly_directories_and_files(self):
+        d = self.make_temp_dir()
+        sd = os.path.join(d, 'subdir')
+        fn = os.path.join(sd, 'file.txt')
+        os.mkdir(sd)
+        open(fn, 'w').close()
+        os.chmod(fn, 0o444)  # readonly
+        os.chmod(sd, 0o444)  # a bad mode for a directory, oops
+        rmtree(sd)
+        assert not os.path.exists(sd)
+
     def test_copy_files(self):
         from check_manifest import copy_files
         actions = []
@@ -208,11 +249,12 @@ class Tests(unittest.TestCase):
     def test_detect_vcs_no_vcs(self):
         from check_manifest import detect_vcs, Failure
         with mock.patch('check_manifest.VCS.detect', staticmethod(lambda *a: False)):
-            with self.assertRaises(Failure) as cm:
-                detect_vcs()
-            self.assertEqual(str(cm.exception),
-                             "Couldn't find version control data"
-                             " (git/hg/bzr/svn supported)")
+            with mock.patch('check_manifest.Git.detect', staticmethod(lambda *a: False)):
+                with self.assertRaises(Failure) as cm:
+                    detect_vcs()
+                self.assertEqual(str(cm.exception),
+                                 "Couldn't find version control data"
+                                 " (git/hg/bzr/svn supported)")
 
     def test_normalize_names(self):
         from check_manifest import normalize_names
@@ -251,6 +293,7 @@ class Tests(unittest.TestCase):
             'setup.cfg',
             'README.txt',
             'src',
+            'src/.gitignore',
             'src/zope',
             'src/zope/__init__.py',
             'src/zope/foo',
@@ -275,7 +318,7 @@ class Tests(unittest.TestCase):
     def test_strip_sdist_extras_with_manifest(self):
         import check_manifest
         from check_manifest import strip_sdist_extras
-        from check_manifest import _get_ignore_from_manifest as parse
+        from check_manifest import _get_ignore_from_manifest_lines as parse
         orig_ignore = check_manifest.IGNORE[:]
         orig_ignore_regexps = check_manifest.IGNORE_REGEXPS[:]
         manifest_in = textwrap.dedent("""
@@ -324,7 +367,7 @@ class Tests(unittest.TestCase):
         # This will change the definitions.
         try:
             # This is normally done in read_manifest:
-            ignore, ignore_regexps = parse(manifest_in)
+            ignore, ignore_regexps = parse(manifest_in.splitlines())
             check_manifest.IGNORE.extend(ignore)
             check_manifest.IGNORE_REGEXPS.extend(ignore_regexps)
             # Filter the file list.
@@ -408,64 +451,70 @@ class Tests(unittest.TestCase):
 
     def test_glob_to_regexp(self):
         from check_manifest import _glob_to_regexp as g2r
-        sep = r'\\' if os.path.sep == '\\' else os.path.sep
-        self.assertEqual(g2r('foo.py'), r'foo\.py\Z(?ms)')
-        self.assertEqual(g2r('foo/bar.py'), r'foo\/bar\.py\Z(?ms)')
-        self.assertEqual(g2r('foo*.py'), r'foo[^%s]*\.py\Z(?ms)' % sep)
-        self.assertEqual(g2r('foo?.py'), r'foo[^%s]\.py\Z(?ms)' % sep)
-        self.assertEqual(g2r('foo[123].py'), r'foo[123]\.py\Z(?ms)')
-        self.assertEqual(g2r('foo[!123].py'), r'foo[^123]\.py\Z(?ms)')
-        self.assertEqual(g2r('foo/*.py'), r'foo\/[^%s]*\.py\Z(?ms)' % sep)
+        sep = os.path.sep.replace('\\', '\\\\')
+        if sys.version_info >= (3, 6):
+            self.assertEqual(g2r('foo.py'), r'(?s:foo\.py)\Z')
+            self.assertEqual(g2r('foo/bar.py'), r'(?s:foo\/bar\.py)\Z')
+            self.assertEqual(g2r('foo*.py'), r'(?s:foo[^%s]*\.py)\Z' % sep)
+            self.assertEqual(g2r('foo?.py'), r'(?s:foo[^%s]\.py)\Z' % sep)
+            self.assertEqual(g2r('foo[123].py'), r'(?s:foo[123]\.py)\Z')
+            self.assertEqual(g2r('foo[!123].py'), r'(?s:foo[^123]\.py)\Z')
+            self.assertEqual(g2r('foo/*.py'), r'(?s:foo\/[^%s]*\.py)\Z' % sep)
+        else:
+            self.assertEqual(g2r('foo.py'), r'foo\.py\Z(?ms)')
+            self.assertEqual(g2r('foo/bar.py'), r'foo\/bar\.py\Z(?ms)')
+            self.assertEqual(g2r('foo*.py'), r'foo[^%s]*\.py\Z(?ms)' % sep)
+            self.assertEqual(g2r('foo?.py'), r'foo[^%s]\.py\Z(?ms)' % sep)
+            self.assertEqual(g2r('foo[123].py'), r'foo[123]\.py\Z(?ms)')
+            self.assertEqual(g2r('foo[!123].py'), r'foo[^123]\.py\Z(?ms)')
+            self.assertEqual(g2r('foo/*.py'), r'foo\/[^%s]*\.py\Z(?ms)' % sep)
 
-    def test_get_ignore_from_manifest(self):
-        from check_manifest import _get_ignore_from_manifest as parse
+    def test_get_ignore_from_manifest_lines(self):
+        from check_manifest import _get_ignore_from_manifest_lines as parse
         from check_manifest import _glob_to_regexp as g2r
         j = os.path.join
         # The return value is a tuple with two lists:
         # ([<list of filename ignores>], [<list of regular expressions>])
-        self.assertEqual(parse(''),
+        self.assertEqual(parse([]),
                          ([], []))
-        self.assertEqual(parse('      \n        '),
+        self.assertEqual(parse(['', ' ']),
                          ([], []))
-        self.assertEqual(parse('exclude *.cfg'),
+        self.assertEqual(parse(['exclude *.cfg']),
                          ([], [g2r('*.cfg')]))
-        self.assertEqual(parse('#exclude *.cfg'),
-                         ([], []))
-        self.assertEqual(parse('exclude          *.cfg'),
+        self.assertEqual(parse(['exclude          *.cfg']),
                          ([], [g2r('*.cfg')]))
-        self.assertEqual(parse('\texclude\t*.cfg foo.*   bar.txt'),
+        self.assertEqual(parse(['\texclude\t*.cfg foo.*   bar.txt']),
                          (['bar.txt'], [g2r('*.cfg'), g2r('foo.*')]))
-        self.assertEqual(parse('exclude some/directory/*.cfg'),
+        self.assertEqual(parse(['exclude some/directory/*.cfg']),
                          ([], [g2r('some/directory/*.cfg')]))
-        self.assertEqual(parse('include *.cfg'),
+        self.assertEqual(parse(['include *.cfg']),
                          ([], []))
-        self.assertEqual(parse('global-exclude *.pyc'),
+        self.assertEqual(parse(['global-exclude *.pyc']),
                          (['*.pyc'], []))
-        self.assertEqual(parse('global-exclude *.pyc *.sh'),
+        self.assertEqual(parse(['global-exclude *.pyc *.sh']),
                          (['*.pyc', '*.sh'], []))
-        self.assertEqual(parse('recursive-exclude dir *.pyc'),
+        self.assertEqual(parse(['recursive-exclude dir *.pyc']),
                          ([j('dir', '*.pyc')], []))
-        self.assertEqual(parse('recursive-exclude dir *.pyc foo*.sh'),
+        self.assertEqual(parse(['recursive-exclude dir *.pyc foo*.sh']),
                          ([j('dir', '*.pyc'), j('dir', 'foo*.sh'),
                            j('dir', '*', 'foo*.sh')], []))
-        self.assertEqual(parse('recursive-exclude dir nopattern.xml'),
+        self.assertEqual(parse(['recursive-exclude dir nopattern.xml']),
                          ([j('dir', 'nopattern.xml'),
                            j('dir', '*', 'nopattern.xml')], []))
         # We should not fail when a recursive-exclude line is wrong:
-        self.assertEqual(parse('recursive-exclude dirwithoutpattern'),
+        self.assertEqual(parse(['recursive-exclude dirwithoutpattern']),
                          ([], []))
-        self.assertEqual(parse('prune dir'),
+        self.assertEqual(parse(['prune dir']),
                          (['dir', j('dir', '*')], []))
         # You should not add a slash at the end of a prune, but let's
         # not fail over it or end up with double slashes.
-        self.assertEqual(parse('prune dir/'),
+        self.assertEqual(parse(['prune dir/']),
                          (['dir', j('dir', '*')], []))
         # You should also not have a leading slash
-        self.assertEqual(parse('prune /dir'),
+        self.assertEqual(parse(['prune /dir']),
                          (['/dir', j('/dir', '*')], []))
         # And a mongo test case of everything at the end
         text = textwrap.dedent("""
-            #exclude *.01
             exclude *.02
             exclude *.03 04.*   bar.txt
             exclude          *.05
@@ -476,7 +525,7 @@ class Tests(unittest.TestCase):
             prune 30
             recursive-exclude    40      *.41
             recursive-exclude 42 *.43 44.*
-        """)
+        """).splitlines()
         self.assertEqual(
             parse(text),
             ([
@@ -498,6 +547,32 @@ class Tests(unittest.TestCase):
                 g2r('some/directory/*.cfg'),
             ]))
 
+    def test_get_ignore_from_manifest(self):
+        from check_manifest import _get_ignore_from_manifest as parse
+        filename = os.path.join(self.make_temp_dir(), 'MANIFEST.in')
+        self.create_file(filename, textwrap.dedent('''
+           exclude \\
+              # yes, this is allowed!
+              test.dat
+
+           # https://github.com/mgedmin/check-manifest/issues/66
+           # docs/ folder
+        '''))
+        self.assertEqual(parse(filename), (['test.dat'], []))
+        self.assertEqual(self.warnings, [])
+
+    def test_get_ignore_from_manifest_warnings(self):
+        from check_manifest import _get_ignore_from_manifest as parse
+        filename = os.path.join(self.make_temp_dir(), 'MANIFEST.in')
+        self.create_file(filename, textwrap.dedent('''
+           # this is bad: a file should not end with a backslash
+           exclude test.dat \\
+        '''))
+        self.assertEqual(parse(filename), (['test.dat'], []))
+        self.assertEqual(self.warnings, [
+            "%s, line 2: continuation line immediately precedes end-of-file" % filename,
+        ])
+
 
 class TestConfiguration(unittest.TestCase):
 
@@ -508,15 +583,18 @@ class TestConfiguration(unittest.TestCase):
         os.chdir(self.tmpdir)
         self.OLD_IGNORE = check_manifest.IGNORE
         self.OLD_IGNORE_REGEXPS = check_manifest.IGNORE_REGEXPS
+        self.OLD_IGNORE_BAD_IDEAS = check_manifest.IGNORE_BAD_IDEAS
         check_manifest.IGNORE = ['default-ignore-rules']
         check_manifest.IGNORE_REGEXPS = ['default-ignore-regexps']
+        check_manifest.IGNORE_BAD_IDEAS = []
 
     def tearDown(self):
         import check_manifest
         check_manifest.IGNORE = self.OLD_IGNORE
         check_manifest.IGNORE_REGEXPS = self.OLD_IGNORE_REGEXPS
+        check_manifest.IGNORE_BAD_IDEAS = self.OLD_IGNORE_BAD_IDEAS
         os.chdir(self.oldpwd)
-        shutil.rmtree(self.tmpdir)
+        rmtree(self.tmpdir)
 
     def test_read_config_no_config(self):
         import check_manifest
@@ -554,6 +632,16 @@ class TestConfiguration(unittest.TestCase):
         self.assertEqual(check_manifest.IGNORE,
                          ['foo', 'bar'])
 
+    def test_read_config_ignore_bad_ideas(self):
+        import check_manifest
+        with open('setup.cfg', 'w') as f:
+            f.write('[check-manifest]\n'
+                    'ignore-bad-ideas = \n'
+                    '  foo\n'
+                    '  bar\n')
+        check_manifest.read_config()
+        self.assertEqual(check_manifest.IGNORE_BAD_IDEAS, ['foo', 'bar'])
+
     def test_read_manifest_no_manifest(self):
         import check_manifest
         check_manifest.read_manifest()
@@ -586,13 +674,16 @@ class TestMain(unittest.TestCase):
         sys.argv = ['check-manifest']
         self.OLD_IGNORE = check_manifest.IGNORE
         self.OLD_IGNORE_REGEXPS = check_manifest.IGNORE_REGEXPS
+        self.OLD_IGNORE_BAD_IDEAS = check_manifest.IGNORE_BAD_IDEAS
         check_manifest.IGNORE = ['default-ignore-rules']
         check_manifest.IGNORE_REGEXPS = ['default-ignore-regexps']
+        check_manifest.IGNORE_BAD_IDEAS = []
 
     def tearDown(self):
         import check_manifest
         check_manifest.IGNORE = self.OLD_IGNORE
         check_manifest.IGNORE_REGEXPS = self.OLD_IGNORE_REGEXPS
+        check_manifest.IGNORE_BAD_IDEAS = self.OLD_IGNORE_BAD_IDEAS
         sys.argv = self._orig_sys_argv
         self._se_patcher.stop()
         self._cm_patcher.stop()
@@ -623,6 +714,13 @@ class TestMain(unittest.TestCase):
         self.assertEqual(check_manifest.IGNORE,
                          ['default-ignore-rules', 'x', 'y', 'z'])
 
+    def test_ignore_bad_ideas_args(self):
+        import check_manifest
+        sys.argv.append('--ignore-bad-ideas=x,y,z')
+        check_manifest.main()
+        self.assertEqual(check_manifest.IGNORE_BAD_IDEAS,
+                         ['x', 'y', 'z'])
+
 
 class TestZestIntegration(unittest.TestCase):
 
@@ -715,6 +813,18 @@ class TestZestIntegration(unittest.TestCase):
 
 class VCSHelper(object):
 
+    command = None  # override in subclasses
+
+    def is_installed(self):
+        try:
+            p = subprocess.Popen([self.command, '--version'],
+                                 stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
+            stdout, stderr = p.communicate()
+            rc = p.wait()
+            return (rc == 0)
+        except OSError:
+            return False
+
     def _run(self, *command):
         # Windows doesn't like Unicode arguments to subprocess.Popen(), on Py2:
         # https://github.com/mgedmin/check-manifest/issues/23#issuecomment-33933031
@@ -733,6 +843,8 @@ class VCSHelper(object):
 class VCSMixin(object):
 
     def setUp(self):
+        if not self.vcs.is_installed() and CAN_SKIP_TESTS:
+            self.skipTest("%s is not installed" % self.vcs.command)
         self.tmpdir = tempfile.mkdtemp(prefix='test-', suffix='-check-manifest')
         self.olddir = os.getcwd()
         os.chdir(self.tmpdir)
@@ -786,6 +898,17 @@ class VCSMixin(object):
                          ['a.txt', 'b', j('b', 'b.txt'), j('b', 'c'),
                           j('b', 'c', 'd.txt')])
 
+    def test_get_vcs_files_deleted_but_not_removed(self):
+        if self.vcs.command == 'bzr':
+            self.skipTest("this cosmetic feature is not supported with bzr")
+            # see the longer explanation in test_missing_source_files
+        from check_manifest import get_vcs_files
+        self._init_vcs()
+        self._create_and_add_to_vcs(['a.txt'])
+        self._commit()
+        os.unlink('a.txt')
+        self.assertEqual(get_vcs_files(), ['a.txt'])
+
     def test_get_vcs_files_in_a_subdir(self):
         from check_manifest import get_vcs_files
         self._init_vcs()
@@ -806,16 +929,26 @@ class VCSMixin(object):
         self._create_and_add_to_vcs([filename])
         self.assertEqual(get_vcs_files(), [filename])
 
+    def test_get_vcs_files_empty(self):
+        from check_manifest import get_vcs_files
+        self._init_vcs()
+        self.assertEqual(get_vcs_files(), [])
+
 
 class GitHelper(VCSHelper):
 
+    command = 'git'
+
     def _init_vcs(self):
         self._run('git', 'init')
         self._run('git', 'config', 'user.name', 'Unit Test')
         self._run('git', 'config', 'user.email', 'test@example.com')
 
     def _add_to_vcs(self, filenames):
-        self._run('git', 'add', '--', *filenames)
+        # Note that we use --force to prevent errors when we want to
+        # add foo.egg-info and the user running the tests has
+        # '*.egg-info' in her global .gitignore file.
+        self._run('git', 'add', '--force', '--', *filenames)
 
     def _commit(self):
         self._run('git', 'commit', '-m', 'Initial')
@@ -824,9 +957,64 @@ class GitHelper(VCSHelper):
 class TestGit(VCSMixin, unittest.TestCase):
     vcs = GitHelper()
 
+    def _init_repo_with_files(self, dirname, filenames):
+        os.mkdir(dirname)
+        os.chdir(dirname)
+        self._init_vcs()
+        self._create_and_add_to_vcs(filenames)
+        self._commit()
+        os.chdir(self.tmpdir)
+
+    def _add_submodule(self, repo, subdir, subrepo):
+        os.chdir(repo)
+        self.vcs._run('git', 'submodule', 'add', subrepo, subdir)
+        self._commit()
+        os.chdir(self.tmpdir)
+
+    def test_detect_git_submodule(self):
+        from check_manifest import detect_vcs, Failure
+        with self.assertRaises(Failure) as cm:
+            detect_vcs()
+        self.assertEqual(str(cm.exception),
+                         "Couldn't find version control data"
+                         " (git/hg/bzr/svn supported)")
+        # now create a .git file like in a submodule
+        open(os.path.join(self.tmpdir, '.git'), 'w').close()
+        self.assertEqual(detect_vcs().metadata_name, '.git')
+
+    def test_get_versioned_files_with_git_submodules(self):
+        from check_manifest import get_vcs_files
+        self._init_repo_with_files('repo1', ['file1', 'file2'])
+        self._init_repo_with_files('repo2', ['file3'])
+        self._init_repo_with_files('repo3', ['file4'])
+        self._add_submodule('repo2', 'sub3', '../repo3')
+        self._init_repo_with_files('main', ['file5', 'subdir/file6'])
+        self._add_submodule('main', 'sub1', '../repo1')
+        self._add_submodule('main', 'subdir/sub2', '../repo2')
+        os.chdir('main')
+        self.vcs._run('git', 'submodule', 'update', '--init', '--recursive')
+        self.assertEqual(
+            get_vcs_files(),
+            [fn.replace('/', os.path.sep) for fn in [
+                '.gitmodules',
+                'file5',
+                'sub1',
+                'sub1/file1',
+                'sub1/file2',
+                'subdir',
+                'subdir/file6',
+                'subdir/sub2',
+                'subdir/sub2/.gitmodules',
+                'subdir/sub2/file3',
+                'subdir/sub2/sub3',
+                'subdir/sub2/sub3/file4',
+            ]])
+
 
 class BzrHelper(VCSHelper):
 
+    command = 'bzr'
+
     def _init_vcs(self):
         self._run('bzr', 'init')
         self._run('bzr', 'whoami', '--branch', 'Unit Test <test@example.com>')
@@ -842,8 +1030,52 @@ class TestBzr(VCSMixin, unittest.TestCase):
     vcs = BzrHelper()
 
 
+@unittest.skipIf(HAS_OEM_CODEC,
+                 "Python 3.6 lets us use 'oem' codec instead of guessing")
+class TestBzrTerminalCharsetDetectionOnOldPythons(unittest.TestCase):
+
+    @mock.patch('sys.stdin')
+    @mock.patch('sys.stdout')
+    def test_terminal_encoding_not_known(self, mock_stdout, mock_stdin):
+        from check_manifest import Bazaar
+        mock_stdout.encoding = None
+        mock_stdin.encoding = None
+        self.assertEqual(Bazaar._get_terminal_encoding(), None)
+
+    @mock.patch('sys.stdout')
+    def test_terminal_encoding_stdout_known(self, mock_stdout):
+        from check_manifest import Bazaar
+        mock_stdout.encoding = 'UTF-8'
+        self.assertEqual(Bazaar._get_terminal_encoding(), 'UTF-8')
+
+    @mock.patch('sys.stdin')
+    @mock.patch('sys.stdout')
+    def test_terminal_encoding_stdin_known(self, mock_stdout, mock_stdin):
+        from check_manifest import Bazaar
+        mock_stdout.encoding = None
+        mock_stdin.encoding = 'UTF-8'
+        self.assertEqual(Bazaar._get_terminal_encoding(), 'UTF-8')
+
+    @mock.patch('sys.stdout')
+    def test_terminal_encoding_cp0(self, mock_stdout):
+        from check_manifest import Bazaar
+        mock_stdout.encoding = 'cp0'
+        self.assertEqual(Bazaar._get_terminal_encoding(), None)
+
+
+@unittest.skipIf(not HAS_OEM_CODEC,
+                 "'oem' codec not available on Python before 3.6")
+class TestBzrTerminalCharsetDetectionOnNewPythons(unittest.TestCase):
+
+    def test_terminal_encoding_cp0(self):
+        from check_manifest import Bazaar
+        self.assertEqual(Bazaar._get_terminal_encoding(), "oem")
+
+
 class HgHelper(VCSHelper):
 
+    command = 'hg'
+
     def _init_vcs(self):
         self._run('hg', 'init')
         with open('.hg/hgrc', 'a') as f:
@@ -862,6 +1094,8 @@ class TestHg(VCSMixin, unittest.TestCase):
 
 class SvnHelper(VCSHelper):
 
+    command = 'svn'
+
     def _init_vcs(self):
         self._run('svnadmin', 'create', 'repo')
         self._run('svn', 'co', 'file:///' + os.path.abspath('repo').replace(os.path.sep, '/'), 'checkout')
@@ -878,8 +1112,21 @@ class SvnHelper(VCSHelper):
 class TestSvn(VCSMixin, unittest.TestCase):
     vcs = SvnHelper()
 
+    def test_svn_externals(self):
+        from check_manifest import get_vcs_files
+        self.vcs._run('svnadmin', 'create', 'repo2')
+        repo2_url = 'file:///' + os.path.abspath('repo2').replace(os.path.sep, '/')
+        self.vcs._init_vcs()
+        self.vcs._run('svn', 'propset', 'svn:externals', 'ext %s' % repo2_url, '.')
+        self.vcs._run('svn', 'up')
+        self._create_files(['a.txt', 'ext/b.txt'])
+        self.vcs._run('svn', 'add', 'a.txt', 'ext/b.txt')
+        j = os.path.join
+        self.assertEqual(get_vcs_files(),
+                         ['a.txt', 'ext', j('ext', 'b.txt')])
 
-class TestUserInterface(unittest.TestCase):
+
+class UIMixin(object):
 
     def setUp(self):
         import check_manifest
@@ -895,6 +1142,21 @@ class TestUserInterface(unittest.TestCase):
         sys.stdout = self.real_stdout
         check_manifest.VERBOSE = self.old_VERBOSE
 
+
+class TestSvnExtraErrors(UIMixin, unittest.TestCase):
+
+    def test_svn_xml_parsing_warning(self):
+        from check_manifest import Subversion
+        entry = ET.XML('<entry path="foo/bar.txt"></entry>')
+        self.assertFalse(Subversion.is_interesting(entry))
+        self.assertEqual(
+            sys.stderr.getvalue(),
+            'svn status --xml parse error: <entry path="foo/bar.txt">'
+            ' without <wc-status>\n')
+
+
+class TestUserInterface(UIMixin, unittest.TestCase):
+
     def test_info(self):
         import check_manifest
         check_manifest.VERBOSE = False
@@ -993,11 +1255,29 @@ class TestUserInterface(unittest.TestCase):
             "Forgot to turn the gas off!\n")
 
 
+def pick_installed_vcs():
+    preferred_order = [GitHelper, HgHelper, BzrHelper, SvnHelper]
+    force = os.getenv('FORCE_TEST_VCS')
+    if force:
+        for cls in preferred_order:
+            if force == cls.command:
+                return cls()
+        raise ValueError('Unsupported FORCE_TEST_VCS=%s (supported: %s)'
+                         % (force, '/'.join(cls.command for cls in preferred_order)))
+    for cls in preferred_order:
+        vcs = cls()
+        if vcs.is_installed():
+            return vcs
+    return None
+
+
 class TestCheckManifest(unittest.TestCase):
 
-    _vcs = GitHelper()
+    _vcs = pick_installed_vcs()
 
     def setUp(self):
+        if self._vcs is None:
+            self.fail('at least one version control system should be installed')
         self.oldpwd = os.getcwd()
         self.tmpdir = tempfile.mkdtemp(prefix='test-', suffix='-check-manifest')
         os.chdir(self.tmpdir)
@@ -1012,14 +1292,25 @@ class TestCheckManifest(unittest.TestCase):
         os.chdir(self.oldpwd)
         rmtree(self.tmpdir)
 
-    def _create_repo_with_code(self):
+    def _create_repo_with_code(self, add_to_vcs=True):
         self._vcs._init_vcs()
         with open('setup.py', 'w') as f:
             f.write("from setuptools import setup\n")
             f.write("setup(name='sample', py_modules=['sample'])\n")
         with open('sample.py', 'w') as f:
             f.write("# wow. such code. so amaze\n")
-        self._vcs._add_to_vcs(['setup.py', 'sample.py'])
+        if add_to_vcs:
+            self._vcs._add_to_vcs(['setup.py', 'sample.py'])
+
+    def _create_repo_with_code_in_subdir(self):
+        os.mkdir('subdir')
+        os.chdir('subdir')
+        self._create_repo_with_code()
+        # NB: when self._vcs is SvnHelper, we're actually in
+        # ./subdir/checkout rather than in ./subdir
+        subdir = os.path.basename(os.getcwd())
+        os.chdir(os.pardir)
+        return subdir
 
     def _add_to_vcs(self, filename, content=''):
         if os.path.sep in filename and not os.path.isdir(os.path.dirname(filename)):
@@ -1035,6 +1326,14 @@ class TestCheckManifest(unittest.TestCase):
         self.assertEqual(str(cm.exception),
                          "This is not a Python project (no setup.py).")
 
+    def test_forgot_to_git_add_anything(self):
+        from check_manifest import check_manifest, Failure
+        self._create_repo_with_code(add_to_vcs=False)
+        with self.assertRaises(Failure) as cm:
+            check_manifest()
+        self.assertEqual(str(cm.exception),
+                         "There are no files added to version control!")
+
     def test_all_is_well(self):
         from check_manifest import check_manifest
         self._create_repo_with_code()
@@ -1042,20 +1341,23 @@ class TestCheckManifest(unittest.TestCase):
 
     def test_relative_pathname(self):
         from check_manifest import check_manifest
-        os.mkdir('subdir')
-        os.chdir('subdir')
-        self._create_repo_with_code()
-        os.chdir(os.pardir)
-        self.assertTrue(check_manifest('subdir'))
+        subdir = self._create_repo_with_code_in_subdir()
+        self.assertTrue(check_manifest(subdir))
 
     def test_relative_python(self):
+        # https://github.com/mgedmin/check-manifest/issues/36
         from check_manifest import check_manifest
-        os.mkdir('subdir')
-        os.chdir('subdir')
-        self._create_repo_with_code()
-        os.chdir(os.pardir)
+        subdir = self._create_repo_with_code_in_subdir()
         python = os.path.relpath(sys.executable)
-        self.assertTrue(check_manifest('subdir', python=python))
+        self.assertTrue(check_manifest(subdir, python=python))
+
+    def test_python_from_path(self):
+        # https://github.com/mgedmin/check-manifest/issues/57
+        # NB: this test assumes you have a 'python' executable somewhere
+        # in your path.
+        from check_manifest import check_manifest
+        subdir = self._create_repo_with_code_in_subdir()
+        self.assertTrue(check_manifest(subdir, python='python'))
 
     def test_suggestions(self):
         from check_manifest import check_manifest
@@ -1133,6 +1435,14 @@ class TestCheckManifest(unittest.TestCase):
         self.assertIn("missing from VCS:\n  MANIFEST.in", sys.stderr.getvalue())
         self.assertNotIn("missing from sdist", sys.stderr.getvalue())
 
+    def test_setup_py_does_not_need_to_be_added_to_be_considered(self):
+        from check_manifest import check_manifest
+        self._create_repo_with_code(add_to_vcs=False)
+        self._add_to_vcs('sample.py')
+        self.assertFalse(check_manifest())
+        self.assertIn("missing from VCS:\n  setup.py", sys.stderr.getvalue())
+        self.assertNotIn("missing from sdist", sys.stderr.getvalue())
+
     def test_bad_ideas(self):
         from check_manifest import check_manifest
         self._create_repo_with_code()
@@ -1144,14 +1454,43 @@ class TestCheckManifest(unittest.TestCase):
         self.assertIn("this also applies to the following:\n  moo.mo",
                       sys.stderr.getvalue())
 
+    def test_ignore_bad_ideas(self):
+        from check_manifest import check_manifest
+        self._create_repo_with_code()
+        with open('setup.cfg', 'w') as f:
+            f.write('[check-manifest]\n'
+                    'ignore =\n'
+                    '  subdir/bar.egg-info\n'
+                    'ignore-bad-ideas =\n'
+                    '  *.mo\n'
+                    '  subdir/bar.egg-info\n')
+        self._add_to_vcs('foo.egg-info')
+        self._add_to_vcs('moo.mo')
+        self._add_to_vcs(os.path.join('subdir', 'bar.egg-info'))
+        self.assertFalse(check_manifest())
+        self.assertIn("you have foo.egg-info in source control!",
+                      sys.stderr.getvalue())
+        self.assertNotIn("moo.mo", sys.stderr.getvalue())
+        self.assertNotIn("bar.egg-info", sys.stderr.getvalue())
+
     def test_missing_source_files(self):
+        # https://github.com/mgedmin/check-manifest/issues/32
         from check_manifest import check_manifest
         self._create_repo_with_code()
         self._add_to_vcs('missing.py')
         os.unlink('missing.py')
         check_manifest()
-        self.assertIn("some files listed as being under source control are missing:\n  missing.py",
-                      sys.stderr.getvalue())
+        if self._vcs.command != 'bzr':
+            # 'bzr ls' doesn't list files that were deleted but not
+            # marked for deletion.  'bzr st' does, but it doesn't list
+            # unmodified files.  Importing bzrlib and using the API to
+            # get the file list we need is (a) complicated, (b) opens
+            # the optional dependency can of worms, and (c) not viable
+            # under Python 3 unless we fork off a Python 2 subprocess.
+            # Manually combining 'bzr ls' and 'bzr st' outputs just to
+            # produce a cosmetic warning message seems like overkill.
+            self.assertIn("some files listed as being under source control are missing:\n  missing.py",
+                        sys.stderr.getvalue())
 
 
 def test_suite():
